# Copyright (C) 2024  Loren Eteval <loren.eteval@proton.me>
#
# This file is part of Furious.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from Furious.QtFramework import *
from Furious.QtFramework import gettext as _
from Furious.Library import *
from Furious.Utility import *
from Furious.Widget.UserSubsQTableWidget import *

from PySide6 import QtWidgets
from PySide6.QtGui import *
from PySide6.QtWidgets import *
from PySide6.QtNetwork import *

import uuid
import functools

__all__ = ['UserSubsWindow']

registerAppSettings('SubscriptionWidgetWindowSize')

needTrans = functools.partial(needTransFn, source=__name__)

needTrans(
    'Add Subscription',
    'Enter subscription remark:',
    'Enter subscription URL:',
    'Cancel',
)


class AddSubsDialog(AppQDialog):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.setWindowTitle(_('Add Subscription'))

        self.remarkText = AppQLabel(_('Enter subscription remark:'))
        self.remarkEdit = QLineEdit()

        self.webURLText = AppQLabel(_('Enter subscription URL:'))
        self.webURLEdit = QLineEdit()

        self.dialogBtns = AppQDialogButtonBox(QtCore.Qt.Orientation.Horizontal)
        self.dialogBtns.addButton(_('OK'), AppQDialogButtonBox.ButtonRole.AcceptRole)
        self.dialogBtns.addButton(
            _('Cancel'), AppQDialogButtonBox.ButtonRole.RejectRole
        )
        self.dialogBtns.accepted.connect(self.accept)
        self.dialogBtns.rejected.connect(self.reject)

        layout = QFormLayout()
        layout.addRow(self.remarkText)
        layout.addRow(self.remarkEdit)
        layout.addRow(self.webURLText)
        layout.addRow(self.webURLEdit)
        layout.addRow(self.dialogBtns)
        layout.setFormAlignment(QtCore.Qt.AlignmentFlag.AlignCenter)

        self.setLayout(layout)

    def setWidthAndHeight(self):
        self.setGeometry(100, 100, 456, 150)

    def subsRemark(self):
        return self.remarkEdit.text()

    def subsWebURL(self):
        return self.webURLEdit.text()


needTrans(
    'Edit Subscription',
    'Subscription List',
    'Add',
    'Delete',
)


class UserSubsWindow(AppQMainWindow):
    def __init__(self, *args, **kwargs):
        callback = kwargs.pop('deleteUniqueCallback', None)

        super().__init__(*args, **kwargs)

        self.setWindowModality(QtCore.Qt.WindowModality.WindowModal)
        self.setWindowTitle(_('Edit Subscription'))

        self.userSubsQTableWidget = UserSubsQTableWidget(deleteUniqueCallback=callback)

        self.userSubsTab = AppQTabWidget(self)
        self.userSubsTab.addTab(self.userSubsQTableWidget, _('Subscription List'))

        # Buttons
        self.addButton = AppQPushButton(_('Add'))
        self.addButton.clicked.connect(lambda: self.addSubs())
        self.deleteButton = AppQPushButton(_('Delete'))
        self.deleteButton.clicked.connect(lambda: self.deleteSelectedItem())

        # Button Layout
        self.buttonWidget = QWidget()
        self.buttonWidgetLayout = QGridLayout(parent=self.buttonWidget)
        self.buttonWidgetLayout.addWidget(self.addButton, 0, 0)
        self.buttonWidgetLayout.addWidget(self.deleteButton, 0, 1)

        self.fakeCentralWidget = QWidget()
        self.fakeCentralWidgetLayout = QVBoxLayout(self.fakeCentralWidget)
        self.fakeCentralWidgetLayout.addWidget(self.userSubsTab)
        self.fakeCentralWidgetLayout.addWidget(self.buttonWidget)

        self.setCentralWidget(self.fakeCentralWidget)

    def setWidthAndHeight(self):
        try:
            windowSize = AppSettings.get('SubscriptionWidgetWindowSize').split(',')

            self.setGeometry(100, 100, *list(int(size) for size in windowSize))
        except Exception:
            # Any non-exit exceptions

            self.setGeometry(100, 100, 360 * GOLDEN_RATIO, 360)

    def addSubs(self):
        def handleResultCode(_addSubsDialog, code):
            if code == PySide6LegacyEnumValueWrapper(AppQDialog.DialogCode.Accepted):
                remark = _addSubsDialog.subsRemark()
                webURL = _addSubsDialog.subsWebURL()

                if remark:
                    # Unique id. Used by display and deletion
                    unique = str(uuid.uuid4())

                    self.userSubsQTableWidget.appendNewItem(
                        unique=unique, remark=remark, webURL=webURL
                    )
            else:
                # Do nothing
                pass

        addSubsDialog = AddSubsDialog(parent=self)
        addSubsDialog.finished.connect(
            functools.partial(handleResultCode, addSubsDialog)
        )

        # dummy ref
        setattr(self, '_addSubsDialog', addSubsDialog)

        # Show the MessageBox asynchronously
        addSubsDialog.open()

    def deleteSelectedItem(self):
        self.userSubsQTableWidget.deleteSelectedItem()

    def keyPressEvent(self, event):
        if event.key() == QtCore.Qt.Key.Key_Delete:
            self.deleteSelectedItem()
        else:
            super().keyPressEvent(event)

    def cleanup(self):
        AppSettings.set(
            'SubscriptionWidgetWindowSize',
            f'{self.geometry().width()},{self.geometry().height()}',
        )



# Copyright (C) 2024  Loren Eteval <loren.eteval@proton.me>
#
# This file is part of Furious.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

from Furious.Interface import *
from Furious.PyFramework import *
from Furious.Utility import *
from Furious.Library import *

__all__ = ['UserServers']

registerAppSettings('Configuration')


class UserServers(SupportExitCleanup, StorageFactory):
    # remark, config, subsId. (subsId corresponds to unique in user subscription)
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        def restore():
            try:
                return UJSONEncoder.decode(
                    PyBase64Encoder.decode(AppSettings.get('Configuration'))
                )
            except Exception:
                # Any non-exit exceptions

                return {'model': []}

        self._data = restore()
        self._list = list(
            constructFromAny(model.pop('config', ''), **model)
            for model in self._data['model']
        )

    def sync(self):
        AppSettings.set(
            'Configuration',
            PyBase64Encoder.encode(
                UJSONEncoder.encode(
                    {'model': list(factory.toStorageObject() for factory in self._list)}
                ).encode()
            ),
        )

    def data(self) -> list[ConfigurationFactory]:
        # Shallow copy
        return self._list

    def cleanup(self):
        self.sync()



# Copyright (C) 2024  Loren Eteval <loren.eteval@proton.me>
#
# This file is part of Furious.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from Furious.PyFramework import *
from Furious.QtFramework import *
from Furious.QtFramework import gettext as _
from Furious.Utility import *
from Furious.TrayActions import *

from PySide6 import QtCore
from PySide6.QtWidgets import QSystemTrayIcon

import logging
import platform
import darkdetect

logger = logging.getLogger(__name__)

__all__ = ['SystemTrayIcon']


class SystemTrayIcon(
    QTranslatable,
    SupportConnectedCallback,
    SupportThemeChangedCallback,
    QSystemTrayIcon,
):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        actions = [
            ConnectAction(isTrayAction=True),
            RoutingAction(isTrayAction=True),
            SystemProxyAction(isTrayAction=True),
            AppQSeperator(),
            ImportAction(isTrayAction=True),
            EditConfigurationAction(isTrayAction=True),
            AppQSeperator(),
            LanguageAction(isTrayAction=True),
            SettingsAction(isTrayAction=True),
            AppQSeperator(),
            ExitAction(isTrayAction=True),
        ]

        # Some old version PySide6 does not have setMenu method
        # for QAction. Protect it. Currently only used in SystemTrayIcon
        if hasattr(AppQAction, 'setMenu'):
            logger.info('contextMenu uses setMenu implementation')

            for action in actions:
                if isinstance(action, AppQAction):
                    if hasattr(self, f'{action}'):
                        logger.warning(f'{self} already has action {action}')

                    setattr(self, f'{action}', action)

            self._menu = AppQMenu(*actions)
            self.setContextMenu(self._menu)
        else:
            logger.info('contextMenu uses addMenu implementation')

            self._refs = []
            self._menu = AppQMenu()

            for action in actions:
                if isinstance(action, AppQAction):
                    if hasattr(self, f'{action}'):
                        logger.warning(f'{self} already has action {action}')

                    setattr(self, f'{action}', action)

                    if action._menu is None:
                        self._menu.addAction(action)
                    else:
                        menu = AppQMenu(*action._menu._actions, title=action.text())
                        menu.setIcon(action.icon())

                        self._refs.append(menu)
                        self._menu.addMenu(menu)
                else:
                    self._menu.addSeparator()

            self.setContextMenu(self._menu)

        self.setDisconnectedIcon()
        self.activated.connect(self.handleActivated)

    def bootstrap(self):
        if AppSettings.isStateON_('StartupOnBoot'):
            # Rrefresh startup application location
            StartupOnBoot.on_()

        if AppSettings.isStateON_('Connect'):
            self.ConnectAction.trigger()

    def showMessage(self, message: str, *args, **kwargs):
        if message:
            super().showMessage(_(APPLICATION_NAME), message, *args, **kwargs)

    def setMonochromeIconByTheme(self, theme):
        def switchMonochrome():
            if theme == 'Dark':
                self.setIcon(bootstrapIconWhite('monochrome-rocket-takeoff.svg'))
            else:
                self.setIcon(bootstrapIcon('monochrome-rocket-takeoff.svg'))

        if PLATFORM == 'Linux':
            # linux. Always use white icon
            self.setIcon(bootstrapIconWhite('monochrome-rocket-takeoff.svg'))
        elif PLATFORM == 'Darwin':
            try:
                release, versioninfo, machine = platform.mac_ver()
                majorRelease = int(release.split('.')[0])

                if majorRelease <= 13:
                    switchMonochrome()
                else:
                    # macOS 14+. Always use white icon
                    self.setIcon(bootstrapIconWhite('monochrome-rocket-takeoff.svg'))
            except Exception:
                # Any non-exit exceptions

                # Fall back to switch
                switchMonochrome()
        else:
            switchMonochrome()

    def setMonochromeIcon(self):
        self.setMonochromeIconByTheme(darkdetect.theme())

    def setDisconnectedIcon(self):
        if AppSettings.isStateON_('UseMonochromeTrayIcon'):
            self.setMonochromeIcon()

            return

        if (
            PLATFORM == 'Darwin'
            or isWindows7()
            or (PLATFORM == 'Windows' and darkdetect.theme() == 'Light')
        ):
            # Darker
            self.setIcon(bootstrapIcon('rocket-takeoff-dark.svg'))
        else:
            self.setIcon(bootstrapIcon('rocket-takeoff.svg'))

    def setConnectedIcon(self):
        if AppSettings.isStateON_('UseMonochromeTrayIcon'):
            self.setMonochromeIcon()

            return

        if isAdministrator():
            self.setIcon(bootstrapIcon('rocket-takeoff-admin-connected.svg'))
        else:
            if (
                PLATFORM == 'Darwin'
                or isWindows7()
                or (PLATFORM == 'Windows' and darkdetect.theme() == 'Light')
            ):
                # Darker
                self.setIcon(bootstrapIcon('rocket-takeoff-connected-dark.svg'))
            else:
                self.setIcon(bootstrapIcon('rocket-takeoff-connected.svg'))

    @QtCore.Slot(QSystemTrayIcon.ActivationReason)
    def handleActivated(self, reason):
        if reason == QSystemTrayIcon.ActivationReason.DoubleClick:
            APP().mainWindow.show()

    def setCustomToolTip(self):
        self.setToolTip(f'{_(APPLICATION_NAME)} {APPLICATION_VERSION}')

    def disconnectedCallback(self):
        self.setDisconnectedIcon()

    def connectedCallback(self):
        self.setConnectedIcon()

    def themeChangedCallback(self, theme):
        if AppSettings.isStateON_('UseMonochromeTrayIcon'):
            # 'theme' is not used. Instead, always query current theme
            self.setMonochromeIconByTheme(darkdetect.theme())

            return

        if APP().isSystemTrayConnected():
            self.setConnectedIcon()
        else:
            self.setDisconnectedIcon()

    def retranslate(self):
        # Nothing to do
        pass



# Copyright (C) 2024  Loren Eteval <loren.eteval@proton.me>
#
# This file is part of Furious.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

from Furious.Interface import *
from Furious.QtFramework import *
from Furious.Library import *
from Furious.Utility import *

from typing import Union

import io
import time
import threading
import multiprocessing

__all__ = ['XrayCore']


def startXrayCore(jsonString: str, msgQueue: multiprocessing.Queue):
    try:
        import xray
    except ImportError:
        # Fake running process
        while True:
            time.sleep(1)
    else:
        if versionToValue(xray.__version__) <= versionToValue('1.8.4'):
            redirect = False
        else:
            # Can be redirected
            redirect = True

        if not isPythonw():
            return StdoutRedirectHelper.launch(
                msgQueue, lambda: xray.startFromJSON(jsonString), redirect
            )

        if not redirect:
            return xray.startFromJSON(jsonString)

        def xrayPythonwProduceMsg():
            try:
                jsonObject = UJSONEncoder.decode(jsonString)
            except Exception:
                # Any non-exit exceptions

                xray.startFromJSON(jsonString)
            else:
                loggingPath = []
                fileStreams = []

                for loggingAttr in ['access', 'error']:
                    try:
                        path = jsonObject['log'][loggingAttr]
                    except Exception:
                        # Any non-exit exceptions

                        continue

                    if path not in loggingPath:
                        loggingPath.append(path)

                        try:
                            stream = open(path, 'rb')
                        except Exception:
                            # Any non-exit exceptions

                            pass
                        else:
                            stream.seek(0, io.SEEK_END)

                            fileStreams.append(stream)

                def produceMsg():
                    while True:
                        for file in fileStreams:
                            for line in iter(file.readline, b''):
                                if line and not line.isspace():
                                    try:
                                        msgQueue.put_nowait(
                                            line.decode('utf-8', 'replace')
                                        )
                                    except Exception:
                                        # Any non-exit exceptions

                                        pass

                        time.sleep(CoreProcess.MSG_PRODUCE_THRESHOLD / 1000)

                try:
                    if fileStreams:
                        msgThread = threading.Thread(target=produceMsg, daemon=True)
                        msgThread.start()

                    xray.startFromJSON(jsonString)
                finally:
                    for stream in fileStreams:
                        stream.close()

        xrayPythonwProduceMsg()


class XrayCore(CoreProcess):
    class ExitCode:
        ConfigurationError = 23
        # Windows: 4294967295. Darwin, Linux: 255 (-1)
        ServerStartFailure = 4294967295 if PLATFORM == 'Windows' else 255
        # Windows shutting down
        SystemShuttingDown = 0x40010004

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @staticmethod
    def name() -> str:
        return 'Xray-core'

    @staticmethod
    def version() -> str:
        try:
            import xray

            return xray.__version__
        except Exception:
            # Any non-exit exceptions

            return '0.0.0'

    def startFromArgs(self, jsonString: str, **kwargs) -> bool:
        self.registerCurrentJSONConfig(jsonString)

        return super().start(
            target=startXrayCore, args=(jsonString, self.msgQueue), **kwargs
        )

    def start(self, config: Union[str, dict], **kwargs) -> bool:
        if isinstance(config, str):
            return self.startFromArgs(config, **kwargs)
        elif isinstance(config, ConfigurationFactory):
            return self.startFromArgs(config.toJSONString(), **kwargs)
        elif isinstance(config, dict):
            try:
                jsonString = UJSONEncoder.encode(config)
            except Exception:
                # Any non-exit exceptions

                return False
            else:
                return self.startFromArgs(jsonString, **kwargs)
        else:
            return False



# Copyright (C) 2024  Loren Eteval <loren.eteval@proton.me>
#
# This file is part of Furious.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

from Furious.Interface import *
from Furious.Utility.Constants import *

from typing import Callable

import os
import sys
import logging
import datetime
import operator
import functools
import traceback
import multiprocessing

__all__ = ['AppMainProcess']

logger = logging.getLogger(__name__)

if PLATFORM == 'Windows':
    ProcessContext = multiprocessing
else:
    ProcessContext = multiprocessing.get_context('spawn')


class AppMainProcess(ProcessContext.Process):
    def __init__(self, loaderFn: Callable[[], ApplicationFactory], **kwargs):
        super().__init__(**kwargs)

        self.startupTime = str(datetime.datetime.now()).replace(':', '')
        self.logFileName = f'{self.startupTime}.log'
        self.fileWritten = multiprocessing.Manager().Value('b', False)

        self.appLoaderFn = loaderFn
        self.application = None

    def exceptHook(self, exceptionType, exceptionValue, tb):
        if logger.level < logging.CRITICAL:
            traceback.print_exception(exceptionType, exceptionValue, tb)

        if isinstance(exceptionValue, AssertionError):
            exitcode = ApplicationFactory.ExitCode.AssertionError
        else:
            exitcode = ApplicationFactory.ExitCode.UnknownException

        logger.error(f'stopped with exitcode {exitcode}')

        self.saveCrashLog(exceptionType, exceptionValue, tb)

        if APP() is not None:
            APP().exit(exitcode)
        else:
            sys.exit(exitcode)

    def saveCrashLog(self, exceptionType, exceptionValue, tb):
        try:
            os.mkdir(CRASH_LOG_DIR)
        except FileExistsError:
            # Directory already exists

            pass
        except Exception:
            # Any non-exit exceptions

            return

        try:
            stackLog = functools.reduce(
                operator.add,
                traceback.format_exception(exceptionType, exceptionValue, tb),
            )

            if APP() is None:
                crashLog = f'{stackLog}'
            else:
                crashLog = f'{APP().log()}\n{stackLog}'

            with open(CRASH_LOG_DIR / self.logFileName, 'w', encoding='utf-8') as file:
                file.write(crashLog)

            self.fileWritten.value = True
        except Exception:
            # Any non-exit exceptions

            pass

    def run(self):
        sys.excepthook = self.exceptHook

        self.application = self.appLoaderFn()

        sys.exit(self.application.run())



# Copyright (C) 2024  Loren Eteval <loren.eteval@proton.me>
#
# This file is part of Furious.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

from Furious.Interface import *
from Furious.PyFramework import *
from Furious.QtFramework import *
from Furious.QtFramework import gettext as _
from Furious.Utility import *

from PySide6 import QtCore
from PySide6.QtGui import *
from PySide6.QtWidgets import *

import io
import pyqrcode

__all__ = ['QRCodeWindow']


class QRCodeWindow(SupportImplicitReference, AppQMainWindow):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.setWindowTitle(_(APPLICATION_NAME))
        self.setFixedSize(640, 640)

        self.tabWidget = QTabWidget(self)
        self.tabWidget.setTabsClosable(True)
        self.tabWidget.tabCloseRequested.connect(self.handleTabCloseRequested)

        self.setCentralWidget(self.tabWidget)

    def tabCount(self) -> int:
        return self.tabWidget.count()

    def initTabByIndex(self, indexes):
        self.tabWidget.clear()

        for index in indexes:
            qrdata = io.BytesIO()

            config = AS_UserServers()[index]

            try:
                uri = config.toURI()
            except Exception:
                # Any non-exit exceptions

                uri = ''

            if uri:
                qrcode = pyqrcode.create(uri)
                qrcode.png(qrdata, scale=5)

                pixmap = QPixmap()
                pixmap.loadFromData(qrdata.getvalue(), 'PNG')

                widget = QLabel(parent=self.tabWidget)
                widget.setPixmap(pixmap)

                self.tabWidget.addTab(
                    widget, f'{index + 1} - ' + config.getExtras('remark')
                )

    @QtCore.Slot(int)
    def handleTabCloseRequested(self, index):
        self.tabWidget.removeTab(index)

        if self.tabWidget.count() == 0:
            self.hide()



# Copyright (C) 2024  Loren Eteval <loren.eteval@proton.me>
#
# This file is part of Furious.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from Furious.Utility.Constants import *
from Furious.Utility.AppSettings import *
from Furious.Utility.SystemRuntime import isScriptMode

from PySide6 import QtCore

import os
import sys
import logging
import plistlib

__all__ = ['StartupOnBoot']

logger = logging.getLogger(__name__)


class StartupOnBoot:
    @staticmethod
    def on_():
        def _on_():
            if isScriptMode():
                # Script mode
                logger.info('ignore turn on StartupOnBoot in script mode')

                return True

            if PLATFORM == 'Windows':
                settings = QtCore.QSettings(
                    'HKEY_CURRENT_USER\\SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Run',
                    QtCore.QSettings.Format.NativeFormat,
                )

                appFilePath = sys.argv[0].replace('/', '\\')

                settings.setValue(APPLICATION_NAME, f'\"{appFilePath}\"')

                return True

            if PLATFORM == 'Linux':
                autostartDir = os.path.join(
                    os.path.expanduser('~'), '.config', 'autostart'
                )

                try:
                    os.mkdir(autostartDir)
                except FileExistsError:
                    pass
                except Exception:
                    # Any non-exit exceptions

                    return False

                try:
                    with open(
                        os.path.join(autostartDir, f'{APPLICATION_NAME}.desktop'),
                        'w',
                        encoding='utf-8',
                    ) as file:
                        file.write(
                            f'[Desktop Entry]\n'
                            f'Exec={sys.argv[0]}\n'
                            f'Version={APPLICATION_VERSION}\n'
                            f'Type=Application\n'
                            f'Categories=Development\n'
                            f'Name={APPLICATION_NAME}\n'
                            f'Terminal=false\n'
                            f'X-GNOME-Autostart-enabled=true\n'
                            f'StartupNotify=false\n'
                            f'X-GNOME-Autostart-Delay=10\n'
                            f'X-MATE-Autostart-Delay=10\n'
                            f'X-KDE-autostart-after=panel\n'
                        )
                except Exception:
                    # Any non-exit exceptions

                    return False
                else:
                    return True

            if PLATFORM == 'Darwin':
                autostartDir = os.path.join(
                    os.path.expanduser('~'), 'Library', 'LaunchAgents'
                )

                try:
                    os.mkdir(autostartDir)
                except FileExistsError:
                    pass
                except Exception:
                    # Any non-exit exceptions

                    return False

                try:
                    plistData = {
                        'Label': f'{APPLICATION_MACOS_SIGNATURE}.{APPLICATION_NAME}',
                        'ProgramArguments': [sys.argv[0]],
                        'RunAtLoad': True,
                    }

                    with open(
                        os.path.join(
                            autostartDir,
                            f'{APPLICATION_MACOS_SIGNATURE}.{APPLICATION_NAME}.plist',
                        ),
                        'wb',
                    ) as file:
                        plistlib.dump(plistData, file)
                except Exception:
                    # Any non-exit exceptions

                    return False
                else:
                    return True

        if _on_():
            logger.info('turn on StartupOnBoot success')
        else:
            logger.error('turn on StartupOnBoot failed')

    @staticmethod
    def off():
        def _off():
            if PLATFORM == 'Windows':
                settings = QtCore.QSettings(
                    'HKEY_CURRENT_USER\\SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Run',
                    QtCore.QSettings.Format.NativeFormat,
                )

                settings.remove(APPLICATION_NAME)

                return True

            if PLATFORM == 'Linux':
                autostartDir = os.path.join(
                    os.path.expanduser('~'), '.config', 'autostart'
                )

                try:
                    os.remove(os.path.join(autostartDir, f'{APPLICATION_NAME}.desktop'))
                except FileNotFoundError:
                    return True
                except Exception:
                    # Any non-exit exceptions

                    return False
                else:
                    return True

            if PLATFORM == 'Darwin':
                autostartDir = os.path.join(
                    os.path.expanduser('~'), 'Library', 'LaunchAgents'
                )

                try:
                    os.remove(
                        os.path.join(
                            autostartDir,
                            f'{APPLICATION_MACOS_SIGNATURE}.{APPLICATION_NAME}.plist',
                        )
                    )
                except FileNotFoundError:
                    return True
                except Exception:
                    # Any non-exit exceptions

                    return False
                else:
                    return True

        if _off():
            logger.info('turn off StartupOnBoot success')
        else:
            logger.error('turn off StartupOnBoot failed')



# Copyright (C) 2024  Loren Eteval <loren.eteval@proton.me>
#
# This file is part of Furious.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

from Furious.Interface import *
from Furious.Utility import *
from Furious.Library.Configuration import *

__all__ = ['getEmptyFactory']

RAY_TEMPLATE = constructFromDict(
    {
        'log': {
            'access': '',
            'error': '',
            'loglevel': 'warning',
        },
        'inbounds': [
            {
                'tag': 'socks',
                'port': 10808,
                'listen': '127.0.0.1',
                'protocol': 'socks',
                'sniffing': {
                    'enabled': True,
                    'destOverride': [
                        'http',
                        'tls',
                    ],
                },
                'settings': {
                    'auth': 'noauth',
                    'udp': True,
                    'allowTransparent': False,
                },
            },
            {
                'tag': 'http',
                'port': 10809,
                'listen': '127.0.0.1',
                'protocol': 'http',
                'sniffing': {
                    'enabled': True,
                    'destOverride': [
                        'http',
                        'tls',
                    ],
                },
                'settings': {
                    'auth': 'noauth',
                    'udp': True,
                    'allowTransparent': False,
                },
            },
        ],
        'outbounds': [
            {
                'tag': 'proxy',
                'protocol': '',
                'settings': {},
                'streamSettings': {
                    'network': 'tcp',
                },
                'mux': {
                    'enabled': False,
                    'concurrency': -1,
                },
            },
            {
                'tag': 'direct',
                'protocol': 'freedom',
                'settings': {},
            },
            {
                'tag': 'block',
                'protocol': 'blackhole',
                'settings': {
                    'response': {
                        'type': 'http',
                    }
                },
            },
        ],
        'routing': {},
    }
)

HY1_TEMPLATE = constructFromDict(
    {
        'server': '',
        'protocol': 'udp',
        'socks5': {
            'listen': '127.0.0.1:10808',
        },
        'http': {
            'listen': '127.0.0.1:10809',
        },
    }
)

HY2_TEMPLATE = constructFromDict(
    {
        'server': '',
        'tls': {
            'insecure': False,
        },
        'socks5': {
            'listen': '127.0.0.1:10808',
        },
        'http': {
            'listen': '127.0.0.1:10809',
        },
    }
)


def getEmptyFactory(protocol: str) -> ConfigurationFactory:
    if (
        protocolRepr(protocol) == Protocol.VMess
        or protocolRepr(protocol) == Protocol.VLESS
    ):
        factory = RAY_TEMPLATE.deepcopy()
        factory['outbounds'][0]['protocol'] = protocol.lower()
        factory['outbounds'][0]['settings']['vnext'] = [
            {
                'address': '',
                'port': 0,
                'users': [{'email': PROXY_OUTBOUND_USER_EMAIL}],
            },
        ]

        return factory

    if (
        protocolRepr(protocol) == Protocol.Shadowsocks
        or protocolRepr(protocol) == Protocol.Trojan
    ):
        factory = RAY_TEMPLATE.deepcopy()
        factory['outbounds'][0]['protocol'] = protocol.lower()
        factory['outbounds'][0]['settings']['servers'] = [
            {
                'address': '',
                'port': 0,
                'email': PROXY_OUTBOUND_USER_EMAIL,
            },
        ]

        return factory

    if protocolRepr(protocol) == Protocol.Hysteria1:
        factory = HY1_TEMPLATE.deepcopy()

        return factory
    if protocolRepr(protocol) == Protocol.Hysteria2:
        factory = HY2_TEMPLATE.deepcopy()

        return factory

    return ConfigurationFactory()



# Copyright (C) 2024  Loren Eteval <loren.eteval@proton.me>
#
# This file is part of Furious.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

from Furious.Interface import *

from typing import Any, AnyStr

import json
import ujson
import base64
import pybase64

__all__ = ['JSONEncoder', 'UJSONEncoder', 'Base64Encoder', 'PyBase64Encoder']


class _JSONEncoder(Encoder):
    def encode(self, data: Any, **kwargs) -> str:
        ensure_ascii = kwargs.pop('ensure_ascii', False)

        return json.dumps(data, ensure_ascii=ensure_ascii, **kwargs)

    def decode(self, data: AnyStr, **kwargs) -> Any:
        return json.loads(data, **kwargs)


class _UJSONEncoder(Encoder):
    def encode(self, data: Any, **kwargs) -> str:
        ensure_ascii = kwargs.pop('ensure_ascii', False)
        escape_forward_slashes = kwargs.pop('escape_forward_slashes', False)

        return ujson.dumps(
            data,
            ensure_ascii=ensure_ascii,
            escape_forward_slashes=escape_forward_slashes,
            **kwargs,
        )

    def decode(self, data: AnyStr, **kwargs) -> Any:
        return ujson.loads(data, **kwargs)


class _Base64Encoder(Encoder):
    def encode(self, data: Any, **kwargs) -> bytes:
        return base64.b64encode(data, **kwargs)

    def decode(self, data: Any, **kwargs) -> bytes:
        validate = kwargs.pop('validate', False)

        return base64.b64decode(data, validate=validate, **kwargs)


class _PyBase64Encoder(Encoder):
    def encode(self, data: Any, **kwargs) -> bytes:
        return pybase64.b64encode(data, **kwargs)

    def decode(self, data: Any, **kwargs) -> bytes:
        validate = kwargs.pop('validate', False)

        return pybase64.b64decode(data, validate=validate, **kwargs)


JSONEncoder = _JSONEncoder()
UJSONEncoder = _UJSONEncoder()
Base64Encoder = _Base64Encoder()
PyBase64Encoder = _PyBase64Encoder()



# Copyright (C) 2024  Loren Eteval <loren.eteval@proton.me>
#
# This file is part of Furious.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

from Furious.Utility.Constants import PLATFORM

from abc import ABC
from typing import Callable, Union

import ujson

__all__ = ['CoreFactory']


class CoreFactory(ABC):
    class ExitCode:
        ConfigurationError = 23
        # Windows: 4294967295. Darwin, Linux: 255 (-1)
        ServerStartFailure = 4294967295 if PLATFORM == 'Windows' else 255
        # Windows shutting down
        SystemShuttingDown = 0x40010004

    def __init__(self, exitCallback: Callable[[CoreFactory, int], None] = None):
        self._exitCallback = exitCallback
        self._jsonConfig = ''

    def registerExitCallback(
        self, exitCallback: Callable[[CoreFactory, int], None]
    ) -> CoreFactory:
        self._exitCallback = exitCallback

        return self

    def registerCurrentJSONConfig(self, config: Union[str, dict]) -> CoreFactory:
        self._jsonConfig = config

        return self

    @staticmethod
    def name() -> str:
        raise NotImplementedError

    @staticmethod
    def version() -> str:
        raise NotImplementedError

    def jsonConfigString(self) -> str:
        if isinstance(self._jsonConfig, str):
            return self._jsonConfig
        if isinstance(self._jsonConfig, dict):
            try:
                return ujson.dumps(
                    self._jsonConfig,
                    ensure_ascii=False,
                    escape_forward_slashes=False,
                )
            except Exception:
                # Any non-exit exceptions

                return ''

        return ''

    def jsonConfigDict(self) -> dict:
        if isinstance(self._jsonConfig, str):
            try:
                return ujson.loads(self._jsonConfig)
            except Exception:
                # Any non-exit exceptions

                return {}
        if isinstance(self._jsonConfig, dict):
            return self._jsonConfig

        return {}

    def start(self, *args, **kwargs) -> bool:
        raise NotImplementedError

    def stop(self):
        raise NotImplementedError



# Copyright (C) 2024  Loren Eteval <loren.eteval@proton.me>
#
# This file is part of Furious.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from Furious.QtFramework import *
from Furious.QtFramework import gettext as _
from Furious.Utility import *

import logging
import functools

__all__ = ['LanguageAction']

logger = logging.getLogger(__name__)

# Handy stuff
SMART_CHOSEN_LANGUAGE = (
    SYSTEM_LANGUAGE if SYSTEM_LANGUAGE in SUPPORTED_LANGUAGE else 'EN'
)

registerAppSettings(
    'Language', validRange=SUPPORTED_LANGUAGE, default=SMART_CHOSEN_LANGUAGE
)

needTrans = functools.partial(needTransFn, source=__name__)


class LanguageChildAction(AppQAction):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def triggeredCallback(self, checked):
        abbr = LANGUAGE_TO_ABBR[self.text()]

        if AppSettings.get('Language') != abbr:
            logger.info(f'set language to \'{self.text()}\'')

            AppSettings.set('Language', abbr)

            QTranslatable.retranslateAll()

    def retranslate(self):
        # Nothing to do
        pass


needTrans('Language')


class LanguageAction(AppQAction):
    def __init__(self, **kwargs):
        super().__init__(
            _('Language'),
            icon=bootstrapIcon('globe2.svg'),
            menu=AppQMenu(
                *list(
                    LanguageChildAction(
                        # Language representation
                        text,
                        checkable=True,
                        checked=text == ABBR_TO_LANGUAGE[AppSettings.get('Language')],
                    )
                    for text in list(LANGUAGE_TO_ABBR.keys())
                ),
            ),
            **kwargs,
        )



# Copyright (C) 2024  Loren Eteval <loren.eteval@proton.me>
#
# This file is part of Furious.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

from Furious.QtFramework import *
from Furious.QtFramework import gettext as _
from Furious.Utility import *

from PySide6.QtWidgets import *

import functools

__all__ = ['IndentSpinBox']

needTrans = functools.partial(needTransFn, source=__name__)

needTrans(
    'Set Indent',
    'Indent:',
    'Cancel',
)


class IndentSpinBox(AppQDialog):
    def __init__(self, parent=None):
        super().__init__(parent)

        self.setWindowTitle(_('Set Indent'))

        self.indentText = AppQLabel(_('Indent:'))

        self.indentSpin = QSpinBox()
        self.indentSpin.setRange(0, 8)
        self.indentSpin.setValue(2)

        self.dialogBtns = AppQDialogButtonBox(QtCore.Qt.Orientation.Horizontal)
        self.dialogBtns.addButton(_('OK'), AppQDialogButtonBox.ButtonRole.AcceptRole)
        self.dialogBtns.addButton(
            _('Cancel'), AppQDialogButtonBox.ButtonRole.RejectRole
        )
        self.dialogBtns.accepted.connect(self.accept)
        self.dialogBtns.rejected.connect(self.reject)

        layout = QFormLayout()
        layout.addRow(self.indentText, self.indentSpin)
        layout.addRow(self.dialogBtns)
        layout.setFormAlignment(QtCore.Qt.AlignmentFlag.AlignCenter)

        self.setLayout(layout)

    def value(self):
        return self.indentSpin.value()



# Copyright (C) 2024  Loren Eteval <loren.eteval@proton.me>
#
# This file is part of Furious.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

from typing import Tuple

import time
import socket
import ipaddress
import functools

__all__ = ['tcping']


def tcping(
    address: str,
    port: int,
    timeout: float,
    count: int,
    interval: float,
) -> Tuple[int, list]:
    host = socket.gethostbyname(address)
    addressobject = ipaddress.ip_address(host)

    if isinstance(addressobject, ipaddress.IPv4Address):
        socketFn = functools.partial(
            socket.socket,
            family=socket.AF_INET,
            type=socket.SOCK_STREAM,
            proto=socket.IPPROTO_TCP,
        )
    elif isinstance(addressobject, ipaddress.IPv6Address):
        socketFn = functools.partial(
            socket.socket,
            family=socket.AF_INET6,
            type=socket.SOCK_STREAM,
            proto=socket.IPPROTO_TCP,
        )
    else:
        raise ValueError('Invalid address')

    sent = 0
    rtts = []

    for sequence in range(count):
        if sequence > 0:
            time.sleep(interval)

        with socketFn() as sock:
            sock.settimeout(timeout)

            counter = time.perf_counter()

            try:
                sock.connect((address, port))
            except Exception:
                # Any non-exit exceptions

                pass
            else:
                sent += 1

                rtts.append(time.perf_counter() - counter)

    return sent, rtts



# Copyright (C) 2024  Loren Eteval <loren.eteval@proton.me>
#
# This file is part of Furious.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

from Furious.Interface import *
from Furious.QtFramework import *
from Furious.QtFramework import gettext as _
from Furious.Utility import *
from Furious.Widget.GuiVTransport import *
from Furious.Widget.GuiVTLS import *

__all__ = ['GuiVLESS']

needTrans = functools.partial(needTransFn, source=__name__)


def getProxyOutboundObject(config: ConfigurationFactory) -> dict:
    if not isinstance(config.get('outbounds'), list):
        config['outbounds'] = []

    for outbound in config['outbounds']:
        if isinstance(outbound, dict):
            tag = outbound.get('tag')

            if isinstance(tag, str) and tag == 'proxy':
                return outbound

    outboundObject = {
        'tag': 'proxy',
        'protocol': 'vless',
        'settings': {
            'vnext': [
                {
                    'address': '',
                    'port': 0,
                    'users': [
                        {
                            'email': PROXY_OUTBOUND_USER_EMAIL,
                        },
                    ],
                },
            ]
        },
        'streamSettings': {
            'network': 'tcp',
        },
        'mux': {
            'enabled': False,
            'concurrency': -1,
        },
    }

    # Proxy outbound not found. Add it
    config['outbounds'].append(outboundObject)

    return outboundObject


def getProxyOutboundServer(config: ConfigurationFactory) -> dict:
    proxyOutbound = getProxyOutboundObject(config)

    try:
        server = proxyOutbound['settings']['vnext'][0]
    except Exception:
        # Any non-exit exceptions

        server = {}

    if not isinstance(server, dict):
        server = {}

    return server


def getProxyOutboundUser(config: ConfigurationFactory) -> dict:
    proxyOutboundServer = getProxyOutboundServer(config)

    try:
        user = proxyOutboundServer['users'][0]
    except Exception:
        # Any non-exit exceptions

        user = {}

    if not isinstance(user, dict):
        user = {}

    return user


class GuiVLESSItemBasicAddress(GuiEditorItemTextInput):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def inputToFactory(self, config: ConfigurationFactory) -> bool:
        proxyOutboundServer = getProxyOutboundServer(config)

        oldAddress = proxyOutboundServer.get('address', '')
        newAddress = self.text()

        if isinstance(oldAddress, str):
            if newAddress != oldAddress:
                if newAddress == '':
                    # Remove field
                    proxyOutboundServer.pop('address', None)
                else:
                    proxyOutboundServer['address'] = newAddress

                return True
            else:
                return False
        else:
            proxyOutboundServer['address'] = newAddress

            return True

    def factoryToInput(self, config: ConfigurationFactory):
        try:
            proxyOutboundServer = getProxyOutboundServer(config)

            self.setText(proxyOutboundServer.get('address', ''))
        except Exception:
            # Any non-exit exceptions

            self.setText('')


class GuiVLESSItemBasicPort(GuiEditorItemTextSpinBox):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Range
        self.setRange(0, 65535)

    def inputToFactory(self, config: ConfigurationFactory) -> bool:
        proxyOutboundServer = getProxyOutboundServer(config)

        oldPort = proxyOutboundServer.get('port')
        newPort = self.value()

        if isinstance(oldPort, int):
            if newPort != oldPort:
                proxyOutboundServer['port'] = newPort

                return True
            else:
                return False
        else:
            proxyOutboundServer['port'] = newPort

            return True

    def factoryToInput(self, config: ConfigurationFactory):
        try:
            proxyOutboundServer = getProxyOutboundServer(config)

            self.setValue(proxyOutboundServer.get('port', 0))
        except Exception:
            # Any non-exit exceptions

            self.setValue(0)


class GuiVLESSItemBasicId(GuiEditorItemTextInput):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def inputToFactory(self, config: ConfigurationFactory) -> bool:
        proxyOutboundUser = getProxyOutboundUser(config)

        oldId = proxyOutboundUser.get('id', '')
        newId = self.text()

        if isinstance(oldId, str):
            if newId != oldId:
                if newId == '':
                    # Remove field
                    proxyOutboundUser.pop('id', None)
                else:
                    proxyOutboundUser['id'] = newId

                return True
            else:
                return False
        else:
            proxyOutboundUser['id'] = newId

            return True

    def factoryToInput(self, config: ConfigurationFactory):
        try:
            proxyOutboundUser = getProxyOutboundUser(config)

            self.setText(proxyOutboundUser.get('id', ''))
        except Exception:
            # Any non-exit exceptions

            self.setText('')


class GuiVLESSItemBasicEncryption(GuiEditorItemTextInput):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def inputToFactory(self, config: ConfigurationFactory) -> bool:
        proxyOutboundUser = getProxyOutboundUser(config)

        oldEncryption = proxyOutboundUser.get('encryption', '')
        newEncryption = self.text()

        if isinstance(oldEncryption, str):
            if newEncryption != oldEncryption:
                if newEncryption == '':
                    # Remove field
                    proxyOutboundUser.pop('encryption', None)
                else:
                    proxyOutboundUser['encryption'] = newEncryption

                return True
            else:
                return False
        else:
            proxyOutboundUser['encryption'] = newEncryption

            return True

    def factoryToInput(self, config: ConfigurationFactory):
        try:
            proxyOutboundUser = getProxyOutboundUser(config)

            self.setText(proxyOutboundUser.get('encryption', ''))
        except Exception:
            # Any non-exit exceptions

            self.setText('')


class GuiVLESSItemBasicFlow(GuiEditorItemTextComboBox):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.addItems(
            [
                '',
                'none',
                'xtls-rprx-vision',
                'xtls-rprx-vision-udp443',
            ]
        )

    def inputToFactory(self, config: ConfigurationFactory) -> bool:
        proxyOutboundUser = getProxyOutboundUser(config)

        oldFlow = proxyOutboundUser.get('flow', '')
        newFlow = self.text()

        if isinstance(oldFlow, str):
            if newFlow != oldFlow:
                proxyOutboundUser['flow'] = newFlow

                return True
            else:
                return False
        else:
            proxyOutboundUser['flow'] = newFlow

            return True

    def factoryToInput(self, config: ConfigurationFactory):
        try:
            proxyOutboundUser = getProxyOutboundUser(config)

            self.setText(proxyOutboundUser.get('flow', ''))
        except Exception:
            # Any non-exit exceptions

            self.setText('')


needTrans(
    'Basic Configuration',
    'Address',
    'Port',
    'Encryption',
    'Flow',
)


class GuiVLESSGroupBoxBasic(GuiEditorWidgetQGroupBox):
    def __init__(self, **kwargs):
        super().__init__(_('Basic Configuration'), **kwargs)

    def containerSequence(self):
        return [
            GuiEditorItemBasicRemark(title=_('Remark')),
            GuiVLESSItemBasicAddress(title=_('Address')),
            GuiVLESSItemBasicPort(title=_('Port')),
            GuiVLESSItemBasicId(title='Id', translatable=False),
            GuiVLESSItemBasicEncryption(title=_('Encryption')),
            GuiVLESSItemBasicFlow(title=_('Flow')),
        ]


needTrans('Proxy')


class GuiVLESSGroupBoxProxy(GuiEditorWidgetQGroupBox):
    def __init__(self, **kwargs):
        super().__init__(_('Proxy'), **kwargs)

    def containerSequence(self):
        return [
            GuiEditorItemProxyHttp(title='http', translatable=False),
            GuiEditorItemProxySocks(title='socks', translatable=False),
        ]


class GuiVLESS(GuiEditorWidgetQDialog):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.setTabText(Protocol.VLESS)

    def groupBoxSequence(self):
        return [
            GuiVLESSGroupBoxBasic(),
            GuiVLESSGroupBoxProxy(),
            GuiVTransportQGroupBox(),
            GuiVTLSQGroupBox(),
        ]



# Copyright (C) 2024  Loren Eteval <loren.eteval@proton.me>
#
# This file is part of Furious.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

from Furious.Interface import *
from Furious.Utility import *

from PySide6 import QtCore

from abc import ABC
from typing import Callable

import os
import sys
import time
import uuid
import logging
import threading
import multiprocessing

__all__ = ['CoreProcess', 'StdoutRedirectHelper']

logger = logging.getLogger(__name__)


class CoreProcess(CoreFactory, ABC):
    MSG_PRODUCE_THRESHOLD = 250

    def __init__(self, **kwargs):
        exitCallback = kwargs.pop('exitCallback', None)

        super().__init__(exitCallback)

        self._process = None
        self._msgQueue = multiprocessing.Queue()
        self._msgCallback = kwargs.pop('msgCallback', None)

        @QtCore.Slot()
        def handleMsgTimemout():
            msg = self.getMsgNoWait()

            if msg and not msg.isspace():
                if callable(self._msgCallback):
                    self._msgCallback(msg)

        self._msgTimer = QtCore.QTimer()
        self._msgTimer.timeout.connect(handleMsgTimemout)

        @QtCore.Slot()
        def handleDaemonTimemout():
            self.checkIsRunning()

        self._daemonTimer = QtCore.QTimer()
        self._daemonTimer.timeout.connect(handleDaemonTimemout)

    @property
    def msgQueue(self) -> multiprocessing.Queue:
        return self._msgQueue

    def registerMsgCallback(self, msgCallback) -> CoreProcess:
        self._msgCallback = msgCallback

        return self

    def isRunning(self) -> bool:
        if isinstance(self._process, multiprocessing.Process):
            return self._process.is_alive()
        else:
            return False

    def checkIsRunning(self) -> bool:
        if isinstance(self._process, multiprocessing.Process):
            if self._process.is_alive():
                return True
            else:
                logger.error(
                    f'{self.name()} stopped unexpectedly with exitcode {self._process.exitcode}'
                )

                self._msgTimer.stop()
                self._daemonTimer.stop()

                if callable(self._exitCallback):
                    self._exitCallback(self, self._process.exitcode)

                # Reset internal process
                self._process = None

                return False
        else:
            return False

    def start(self, **kwargs) -> bool:
        daemon = kwargs.pop('daemon', True)
        waitCore = kwargs.pop('waitCore', True)
        waitTime = kwargs.pop('waitTime', 2500)

        self._process = multiprocessing.Process(**kwargs, daemon=daemon)
        self._process.start()

        logger.info(f'{self.name()} {self.version()} started')

        self._msgTimer.start(self.MSG_PRODUCE_THRESHOLD)

        if waitCore:
            # Wait for the core to start up completely
            PySide6LegacyEventLoopWait(waitTime)

        if self.checkIsRunning():
            # Start core daemon
            self._daemonTimer.start(CORE_CHECK_ALIVE_INTERVAL)

            return True
        else:
            return False

    def stop(self):
        if self.isRunning():
            self._msgTimer.stop()
            self._daemonTimer.stop()

            self._process.terminate()
            self._process.join()

            logger.info(
                f'{self.name()} terminated with exitcode {self._process.exitcode}'
            )

    def getMsgNoWait(self) -> str:
        try:
            return self._msgQueue.get_nowait()
        except Exception:
            # Any non-exit exceptions

            return ''


class StdoutRedirectHelper:
    TemporaryDir = QtCore.QTemporaryDir()

    @staticmethod
    def launch(
        msgQueue: multiprocessing.Queue, entrypoint: Callable[[], None], redirect: bool
    ):
        if not callable(entrypoint):
            return

        if (
            not StdoutRedirectHelper.TemporaryDir.isValid()
            or not redirect
            # pythonw.exe
            or isPythonw()
        ):
            # Call entrypoint directly
            entrypoint()

            return

        temporaryFile = StdoutRedirectHelper.TemporaryDir.filePath(str(uuid.uuid4()))
        tmpFileStream = open(temporaryFile, 'w+b')
        stdoutFileno_ = sys.stdout.fileno()
        stderrFileno_ = sys.stderr.fileno()

        sys.stdout.close()
        sys.stderr.close()

        # Redirect
        os.dup2(tmpFileStream.fileno(), stdoutFileno_)
        os.dup2(tmpFileStream.fileno(), stderrFileno_)

        sys.stdout = tmpFileStream
        sys.stderr = tmpFileStream

        def produceMsg():
            with open(temporaryFile, 'rb') as file:
                while True:
                    for line in iter(file.readline, b''):
                        if line and not line.isspace():
                            try:
                                msgQueue.put_nowait(line.decode('utf-8', 'replace'))
                            except Exception:
                                # Any non-exit exceptions

                                pass

                    time.sleep(CoreProcess.MSG_PRODUCE_THRESHOLD / 1000)

        msgThread = threading.Thread(target=produceMsg, daemon=True)
        msgThread.start()

        with tmpFileStream:
            entrypoint()



# Copyright (C) 2024  Loren Eteval <loren.eteval@proton.me>
#
# This file is part of Furious.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

from Furious.Interface import *
from Furious.QtFramework import *
from Furious.QtFramework import gettext as _
from Furious.Utility import *

from PySide6 import QtCore
from PySide6.QtGui import *
from PySide6.QtWidgets import *

import logging
import functools

__all__ = ['GuiHysteria1']

logger = logging.getLogger(__name__)

needTrans = functools.partial(needTransFn, source=__name__)


class GuiHy1ItemTextInput(GuiEditorItemTextInput):
    def __init__(self, *args, **kwargs):
        key = kwargs.pop('key', '')

        super().__init__(*args, **kwargs)

        self.key = key

    def inputToFactory(self, config: ConfigurationFactory) -> bool:
        oldValue = config.get(self.key, '')
        newValue = self.text()

        if isinstance(oldValue, str):
            if newValue != oldValue:
                if newValue == '':
                    # Remove field
                    config.pop(self.key, None)
                else:
                    config[self.key] = newValue

                return True
            else:
                return False
        else:
            config[self.key] = newValue

            return True

    def factoryToInput(self, config: ConfigurationFactory):
        try:
            self.setText(config.get(self.key, ''))
        except Exception:
            # Any non-exit exceptions

            self.setText('')


class GuiHy1ItemBasicProtocol(GuiEditorItemTextComboBox):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.addItems(['', 'udp', 'wechat-video', 'faketcp'])

    def inputToFactory(self, config: ConfigurationFactory) -> bool:
        oldProtocol = config.get('protocol', '')
        newProtocol = self.text()

        if isinstance(oldProtocol, str):
            if newProtocol != oldProtocol:
                config['protocol'] = newProtocol

                return True
            else:
                return False
        else:
            config['protocol'] = newProtocol

            return True

    def factoryToInput(self, config: ConfigurationFactory):
        try:
            self.setText(config.get('protocol', 'udp'))
        except Exception:
            # Any non-exit exceptions

            self.setText('')


class GuiHy1ItemSpeedUpMbps(GuiEditorItemTextSpinBox):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Range
        self.setRange(0, 1048576)

    def inputToFactory(self, config: ConfigurationFactory) -> bool:
        oldUpMbps = config.get('up_mbps')
        newUpMbps = self.value()

        if isinstance(oldUpMbps, int):
            if newUpMbps != oldUpMbps:
                config['up_mbps'] = newUpMbps

                return True
            else:
                return False
        else:
            config['up_mbps'] = newUpMbps

            return True

    def factoryToInput(self, config: ConfigurationFactory):
        try:
            self.setValue(config.get('up_mbps'))
        except Exception:
            # Any non-exit exceptions

            self.setValue(24)


class GuiHy1ItemSpeedDownMbps(GuiEditorItemTextSpinBox):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Range
        self.setRange(0, 1048576)

    def inputToFactory(self, config: ConfigurationFactory) -> bool:
        oldDownMbps = config.get('down_mbps')
        newDownMbps = self.value()

        if isinstance(oldDownMbps, int):
            if newDownMbps != oldDownMbps:
                config['down_mbps'] = newDownMbps

                return True
            else:
                return False
        else:
            config['down_mbps'] = newDownMbps

            return True

    def factoryToInput(self, config: ConfigurationFactory):
        try:
            self.setValue(config.get('down_mbps'))
        except Exception:
            # Any non-exit exceptions

            self.setValue(96)


class GuiHy1ItemTLSInsecure(GuiEditorItemTextCheckBox):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def inputToFactory(self, config: ConfigurationFactory) -> bool:
        try:
            oldChecked = config.get('insecure')
        except Exception:
            # Any non-exit exceptions

            oldChecked = False

        newChecked = self.isChecked()

        if newChecked:
            if oldChecked is not True:
                config['insecure'] = True

                return True
            else:
                return False
        else:
            if oldChecked is True:
                config['insecure'] = False

                return True
            if oldChecked is False:
                return False

            # Invalid value for insecure. Remove it
            config.pop('insecure', None)

            # Modified silently
            return False

    def factoryToInput(self, config: ConfigurationFactory):
        try:
            checked = config['insecure']
        except Exception:
            # Any non-exit exceptions

            checked = False

        if not isinstance(checked, bool):
            checked = False

        self.setChecked(checked)


needTrans('Project Website')


class GuiHy1ProjectWebsiteURL(AppQLabel):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.setWebsiteURL()
        self.linkActivated.connect(self.handleLinkActivated)

    def setWebsiteURL(self):
        self.setText(
            '<html><head/><body><p>'
            '<a href=\"https://v1.hysteria.network/\">'
            '<span style=\" text-decoration: underline; color:#007ad6;\">'
            + _('Project Website')
            + '</span></a></p></body></html>'
        )

    @staticmethod
    def handleLinkActivated(link: str):
        if QDesktopServices.openUrl(QtCore.QUrl(link)):
            logger.info(f'open link \'{link}\' success')
        else:
            logger.error(f'open link \'{link}\' failed')

    def retranslate(self):
        self.setWebsiteURL()


needTrans(
    'Basic Configuration',
    'Remark',
    'Server',
    'Protocol',
)


class GuiHy1GroupBoxBasic(GuiEditorWidgetQGroupBox):
    def __init__(self, **kwargs):
        super().__init__(_('Basic Configuration'), **kwargs)

    def containerSequence(self):
        return [
            GuiEditorItemBasicRemark(title=_('Remark')),
            GuiHy1ItemTextInput(title=_('Server'), key='server'),
            GuiHy1ItemBasicProtocol(title=_('Protocol')),
            GuiHy1ItemTextInput(title='auth_str', translatable=False, key='auth_str'),
            GuiHy1ItemTextInput(title='obfs', translatable=False, key='obfs'),
        ]


needTrans('Proxy')


class GuiHy1GroupBoxProxy(GuiEditorWidgetQGroupBox):
    def __init__(self, **kwargs):
        super().__init__(_('Proxy'), **kwargs)

    def containerSequence(self):
        return [
            GuiEditorItemProxyHttp(title='http', translatable=False),
            GuiEditorItemProxySocks(title='socks', translatable=False),
        ]


needTrans('Speed')


class GuiHy1GroupBoxSpeed(GuiEditorWidgetQGroupBox):
    def __init__(self, **kwargs):
        super().__init__(_('Speed'), **kwargs)

    def containerSequence(self):
        return [
            GuiHy1ItemSpeedUpMbps(title='up_mbps', translatable=False),
            GuiHy1ItemSpeedDownMbps(title='down_mbps', translatable=False),
        ]


class GuiHy1GroupBoxTLS(GuiEditorWidgetQGroupBox):
    def __init__(self, **kwargs):
        super().__init__('TLS', **kwargs)

        self.translatable = False

    def containerSequence(self):
        return [
            GuiHy1ItemTextInput(title='sni', translatable=False, key='server_name'),
            GuiHy1ItemTextInput(title='alpn', translatable=False, key='alpn'),
            GuiHy1ItemTextInput(title='ca', translatable=False, key='ca'),
            GuiHy1ItemTLSInsecure(title='insecure', translatable=False),
        ]


needTrans('Other')


class GuiHy1GroupBoxOther(GuiEditorItemFactory, AppQGroupBox):
    def __init__(self, **kwargs):
        super().__init__(_('Other'), **kwargs)

        self._website = GuiHy1ProjectWebsiteURL()

        layout = QFormLayout()
        layout.addRow(self._website)

        self.setLayout(layout)

    def inputToFactory(self, config: ConfigurationFactory) -> bool:
        return False

    def factoryToInput(self, config: ConfigurationFactory):
        pass


class GuiHysteria1(GuiEditorWidgetQDialog):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.setTabText(Protocol.Hysteria1)

    def groupBoxSequence(self):
        return [
            GuiHy1GroupBoxBasic(),
            GuiHy1GroupBoxProxy(),
            GuiHy1GroupBoxSpeed(),
            GuiHy1GroupBoxTLS(),
            # GuiHy1GroupBoxOther(),
        ]



# Copyright (C) 2024  Loren Eteval <loren.eteval@proton.me>
#
# This file is part of Furious.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

from Furious.QtFramework.Ancestors import *
from Furious.QtFramework.DynamicTheme import AppHue
from Furious.QtFramework.DynamicTranslate import gettext as _, needTransFn
from Furious.QtFramework.QtGui import AppQAction, AppQSeperator
from Furious.PyFramework import *
from Furious.Utility import *
from Furious.Library import *

from PySide6 import QtCore
from PySide6.QtGui import *
from PySide6.QtWidgets import *

import functools

__all__ = [
    'moveToCenter',
    'AppQCheckBox',
    'AppQDialog',
    'AppQDialogButtonBox',
    'AppQGroupBox',
    'AppQHeaderView',
    'AppQLabel',
    'AppQListWidget',
    'AppQMainWindow',
    'AppQMenu',
    'AppQMenuBar',
    'AppQMessageBox',
    'AppQPushButton',
    'AppQTableWidget',
    'AppQTabWidget',
    'AppQToolBar',
    'QuestionDeleteMBox',
    'NewChangesNextTimeMBox',
]

needTrans = functools.partial(needTransFn, source=__name__)


def moveToCenter(widget, parent=None):
    geometry = widget.geometry()

    if parent is None:
        center = QApplication.primaryScreen().availableGeometry().center()
    else:
        center = parent.geometry().center()

    geometry.moveCenter(center)

    widget.move(geometry.topLeft())


class AppQCheckBox(QTranslatable, QCheckBox):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def retranslate(self):
        self.setText(_(self.text()))


class AppQDialog(QTranslatable, SupportConnectedCallback, QDialog):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._firstShowCall = True

        if PLATFORM != 'Darwin':
            self.setWidthAndHeight()

        self.setWindowIcon(AppHue.currentWindowIcon())

    def setWidthAndHeight(self):
        pass

    def exec(self):
        self.show()

        return super().exec()

    def open(self):
        self.show()

        return super().open()

    def show(self):
        super().show()

        if PLATFORM == 'Darwin':
            if self._firstShowCall:
                APP().processEvents()

                self.setWidthAndHeight()

                self._firstShowCall = False

        moveToCenter(self)

    def retranslate(self):
        self.setWindowTitle(_(self.windowTitle()))

    def disconnectedCallback(self):
        self.setWindowIcon(AppHue.disconnectedWindowIcon())

    def connectedCallback(self):
        self.setWindowIcon(AppHue.connectedWindowIcon())


class AppQDialogButtonBox(QTranslatable, QDialogButtonBox):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def retranslate(self):
        for button in self.buttons():
            button.setText(_(button.text()))


class AppQGroupBox(QTranslatable, QGroupBox):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def retranslate(self):
        self.setTitle(_(self.title()))


class AppQHeaderView(SupportExitCleanup, SupportConnectedCallback, QHeaderView):
    def sectionSizeSettingsEmpty(self):
        return self.sectionSizeSettingsName == ''

    def __init__(self, *args, **kwargs):
        self.sectionSizeSettingsName = kwargs.pop('sectionSizeSettingsName', '')

        super().__init__(*args, **kwargs)

        self.columnCount = self.parent().columnCount()
        self.sectionSizeTable = {}

        self.setSectionsClickable(True)
        self.setStyleSheet(self.getStyleSheet(AppHue.currentColor()))
        self.setFont(QFont(APP().customFontName))

        self.sectionResized.connect(self.handleSectionResized)

    def restoreSectionSize(self):
        if self.sectionSizeSettingsEmpty():
            return

        try:
            # https://bugreports.qt.io/browse/QTBUG-119862
            # Affected: PySide6 6.6.1+
            self.setDefaultSectionSize(self.defaultSectionSize())

            self.sectionSizeTable = UJSONEncoder.decode(
                AppSettings.get(self.sectionSizeSettingsName)
            )

            # Fill missing value
            for column in range(self.columnCount):
                if self.sectionSizeTable.get(str(column)) is None:
                    self.sectionSizeTable[str(column)] = self.defaultSectionSize()

            with QBlockSignals(self):
                for key, value in self.sectionSizeTable.items():
                    self.resizeSection(int(key), value)
        except Exception:
            # Any non-exit exceptions

            # Leave keys as strings since they will be
            # loaded as string from json
            self.sectionSizeTable = {
                str(column): self.defaultSectionSize()
                for column in range(self.columnCount)
            }

    def setCustomSectionResizeMode(self):
        # Horizontal header resize mode
        for index in range(self.columnCount):
            if index < self.columnCount - 1:
                self.setSectionResizeMode(index, AppQHeaderView.ResizeMode.Interactive)
            else:
                self.setSectionResizeMode(index, AppQHeaderView.ResizeMode.Stretch)

    @staticmethod
    def getStyleSheet(color):
        return f'QHeaderView::section:hover {{ background-color: {color}; }}'

    def disconnectedCallback(self):
        self.setStyleSheet(self.getStyleSheet(AppHue.disconnectedColor()))

    def connectedCallback(self):
        self.setStyleSheet(self.getStyleSheet(AppHue.connectedColor()))

    @QtCore.Slot(int, int, int)
    def handleSectionResized(self, index: int, oldSize: int, newSize: int):
        if self.sectionSizeSettingsEmpty():
            return

        # Keys are string when loaded from json
        self.sectionSizeTable[str(index)] = newSize

    def cleanup(self):
        if self.sectionSizeSettingsEmpty():
            return

        AppSettings.set(
            self.sectionSizeSettingsName,
            UJSONEncoder.encode(self.sectionSizeTable),
        )


class AppQLabel(QTranslatable, QLabel):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def retranslate(self):
        self.setText(_(self.text()))


class AppQListWidget(SupportConnectedCallback, QListWidget):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.setSelectionColor(AppHue.disconnectedColor())

    def setSelectionColor(self, color):
        self.setStyleSheet(
            f'QListWidget::item:selected {{'
            f'    background: {color};'
            f'}}'
            f''
            f'QListWidget::item:hover {{'
            f'    background: {color};'
            f'}}'
        )

    @property
    def selectedIndex(self):
        return sorted(list(set(index.row() for index in self.selectedIndexes())))

    def disconnectedCallback(self):
        self.setSelectionColor(AppHue.disconnectedColor())

    def connectedCallback(self):
        self.setSelectionColor(AppHue.connectedColor())


class AppQMainWindow(
    QTranslatable,
    SupportConnectedCallback,
    SupportExitCleanup,
    QMainWindow,
):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._firstShowCall = True

        self.setWindowIcon(AppHue.currentWindowIcon())

        self._menuBar = AppQMenuBar(parent=self)
        self.setMenuBar(self._menuBar)

        if PLATFORM != 'Darwin':
            self.setWidthAndHeight()

    def setWidthAndHeight(self):
        pass

    def show(self):
        super().show()

        if PLATFORM == 'Darwin':
            if self._firstShowCall:
                APP().processEvents()

                self.setWidthAndHeight()

                APP().processEvents()

                self._firstShowCall = False

        moveToCenter(self)

    def closeEvent(self, event):
        event.ignore()

        self.hide()

    def retranslate(self):
        self.setWindowTitle(_(self.windowTitle()))

    def disconnectedCallback(self):
        self.setWindowIcon(AppHue.disconnectedWindowIcon())

    def connectedCallback(self):
        self.setWindowIcon(AppHue.connectedWindowIcon())

    def cleanup(self):
        pass


class AppQMenu(QTranslatable, SupportConnectedCallback, QMenu):
    def __init__(self, *actions, **kwargs):
        super().__init__(**kwargs)

        # In some old version PySide6, the self.actions() method
        # does not return with seperators. _actions list append
        # them all
        self._actions = []

        for action in actions:
            if isinstance(action, AppQSeperator):
                self._actions.append(action)
                self.addSeparator()
            elif isinstance(action, AppQAction):
                self._actions.append(action)
                self.addAction(action)
            else:
                # Do nothing
                pass

        self.setStyleSheet(self.getStyleSheet(AppHue.currentColor()))

    @staticmethod
    def getStyleSheet(color):
        return (
            f'QMenu::item {{'
            f'    background-color: solid;'
            f'}}'
            f''
            f'QMenu::item:selected {{'
            f'    background-color: {color};'
            f'}}'
        )

    def retranslate(self):
        self.setTitle(_(self.title()))

    def disconnectedCallback(self):
        self.setStyleSheet(self.getStyleSheet(AppHue.disconnectedColor()))

    def connectedCallback(self):
        self.setStyleSheet(self.getStyleSheet(AppHue.connectedColor()))


class AppQMenuBar(SupportConnectedCallback, QMenuBar):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.setSelectionColor(AppHue.currentColor())

    @staticmethod
    def getStyleSheet(color):
        return (
            f'QMenuBar::item:selected {{'
            f'    background: {color};'
            f'}}'
            f''
            f'QMenuBar::item:hover {{'
            f'    background: {color};'
            f'}}'
        )

    def setSelectionColor(self, color):
        self.setStyleSheet(self.getStyleSheet(color))

    def disconnectedCallback(self):
        self.setSelectionColor(AppHue.disconnectedColor())

    def connectedCallback(self):
        self.setSelectionColor(AppHue.connectedColor())


class AppQMessageBox(QTranslatable, SupportConnectedCallback, QMessageBox):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.setWindowIcon(AppHue.currentWindowIcon())

    def moveToCenter(self):
        moveToCenter(self, self.parentWidget())

        return self

    def exec(self):
        self.show()
        self.moveToCenter()

        return super().exec()

    def open(self):
        self.show()
        self.moveToCenter()

        return super().open()

    def retranslate(self):
        self.setWindowTitle(_(self.windowTitle()))
        self.setText(_(self.text()))

        try:
            self.setInformativeText(_(self.informativeText()))
        except KeyError:
            # Any translatable informative text
            pass

        for button in self.buttons():
            if button.text().find('OK') != -1:
                # &OK...
                pass
            else:
                button.setText(_(button.text()))

        self.moveToCenter()

    def disconnectedCallback(self):
        self.setWindowIcon(AppHue.disconnectedWindowIcon())

    def connectedCallback(self):
        self.setWindowIcon(AppHue.connectedWindowIcon())


class AppQPushButton(QTranslatable, QPushButton):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def retranslate(self):
        self.setText(_(self.text()))


class AppQTableWidget(SupportConnectedCallback, QTableWidget):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.setWordWrap(False)

    @property
    def selectedIndex(self):
        return sorted(list(set(index.row() for index in self.selectedIndexes())))

    @staticmethod
    def getStyleSheet(color):
        return f'QTableWidget {{ selection-background-color: {color}; }}'

    def setSelectionColor(self, color):
        self.setStyleSheet(self.getStyleSheet(color))

    def activateItemByIndex(self, index, activate):
        if activate:
            for column in range(self.columnCount()):
                item = self.item(int(index), column)

                if item is None:
                    # Do nothing
                    continue

                font = item.font()
                font.setBold(True)

                item.setFont(font)
                item.setForeground(QColor(AppHue.currentColor()))
        else:
            for column in range(self.columnCount()):
                item = self.item(int(index), column)

                if item is None:
                    # Do nothing
                    continue

                font = item.font()
                font.setBold(False)

                item.setFont(font)
                item.setForeground(QBrush())

    def selectMultipleRows(self, indexes: list[int], clearCurrentSelection: bool):
        if clearCurrentSelection:
            self.selectionModel().clearSelection()

        selection = self.selectionModel().selection()

        for index in indexes:
            selection.select(
                self.model().index(index, 0),
                self.model().index(index, self.columnCount() - 1),
            )

        self.selectionModel().select(
            selection, QtCore.QItemSelectionModel.SelectionFlag.Select
        )

    def disconnectedCallback(self):
        self.setSelectionColor(AppHue.disconnectedColor())

    def connectedCallback(self):
        self.setSelectionColor(AppHue.connectedColor())


class AppQTabWidget(QTranslatable, QTabWidget):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def retranslate(self):
        for index in range(self.count()):
            self.setTabText(index, _(self.tabText(index)))


class AppQToolBar(QTranslatable, QToolBar):
    def __init__(self, *actions, **kwargs):
        super().__init__(**kwargs)

        self._actions = []

        for action in actions:
            if isinstance(action, AppQSeperator):
                self._actions.append(action)
                self.addSeparator()
            elif isinstance(action, AppQAction):
                self._actions.append(action)
                self.addAction(action)
            else:
                # Do nothing
                pass

        self.setStyleSheet(self.getStyleSheet())

    @staticmethod
    def getStyleSheet():
        return f'QToolBar {{ spacing: 5px; }}'

    def retranslate(self):
        self.setWindowTitle(_(self.windowTitle()))


needTrans(
    'Delete',
    'Delete these items?',
    'Delete this item?',
)


class QuestionDeleteMBox(AppQMessageBox):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.isMulti = False
        self.possibleRemark = ''

        self.setWindowTitle(_('Delete'))
        self.setStandardButtons(
            AppQMessageBox.StandardButton.Yes | AppQMessageBox.StandardButton.No
        )

    def customText(self) -> str:
        if self.isMulti:
            return _('Delete these items?')
        else:
            return _('Delete this item?') + f'\n\n{self.possibleRemark}'

    def retranslate(self):
        self.setWindowTitle(_(self.windowTitle()))
        self.setText(self.customText())

        # Ignore informative text, buttons

        self.moveToCenter()


needTrans('New changes will take effect next time')


class NewChangesNextTimeMBox(AppQMessageBox):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.setIcon(AppQMessageBox.Icon.Information)
        self.setText(_('New changes will take effect next time'))
